//
// Written by Igor Tolmachev, IT Samples
//        mailto:support@itsamples.com
//
// Copyright (c) 2009.
//
// This code may be used in compiled form in any way you desire. This
// file may be redistributed unmodified by any means PROVIDING it is 
// not sold for profit without the authors written consent, and 
// providing that this notice and the authors name and all copyright 
// notices remains intact. If the source code in this file is used in 
// any  commercial application then a statement along the lines of 
// "Portions copyright (c) Igor Tolmachev, 2009" must be included in 
// the startup banner, "About" box or printed documentation. An email 
// letting me know that you are using it would be nice as well. That's 
// not much to ask considering the amount of work that went into this.
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability for any damage/loss of business that
// this product may cause.
//
// Expect bugs!
// 
// Please use and enjoy, and let me know of any bugs/mods/improvements 
// that you have found/implemented and I will fix/incorporate them into 
// this file. 


// MainFrm.cpp : implementation of the CMainFrame class
//

#include "stdafx.h"
#include "NetworkLights.h"
#include "AboutDialog.h"
#include "StatisticsDialog.h"
#include "SettingsDialog.h"
#include "MainFrm.h"
#include "WinAble.h"

static BOOL m_bContinue = TRUE;
static UINT TxThread(LPVOID Info);  
static UINT RxThread(LPVOID Info);  

/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNAMIC(CMainFrame, CNewFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CNewFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_MESSAGE(WM_ICONTRAY,OnIconTray) 
	ON_WM_DESTROY()
	ON_COMMAND(ID_ABOUT, OnAbout)
	ON_COMMAND(ID_SETTINGS, OnSettings)
	ON_COMMAND(ID_STATISTICS, OnStatistics)
	ON_COMMAND(ID_EXIT, OnExit)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
	m_nDuration = 100;
	m_pRxThread = NULL;
	m_pTxThread = NULL;
	m_bRxContinue = TRUE;
	m_bTxContinue = TRUE;
	m_bIsForced = FALSE;
}

CMainFrame::~CMainFrame()
{
}

/////////////////////////////////////////////////////////////////////////////

UINT TxThread(LPVOID pInfo)
{
	CMainFrame* ptr = (CMainFrame*)pInfo;
	UINT nSegRcvd = 0;
	UINT nSegSent = 0;
	MIB_TCPSTATS  mibTcpStats;
	UINT nCounter = 0;
	DWORD dwLocalOutSegs = 0;

	while(m_bContinue)
	{
		if(ptr->m_bTxContinue)
		{
			if(GetTcpStatistics(&mibTcpStats) == NO_ERROR)
			{
				dwLocalOutSegs = mibTcpStats.dwOutSegs;
				if(dwLocalOutSegs>nSegSent)
				{
					nSegSent = dwLocalOutSegs;
					ptr->FlashNumLock();
				}
				else if(dwLocalOutSegs==nSegRcvd) nCounter++;

				if(nCounter==15)
				{
					nCounter=0;
					ptr->ResetIndicators();
				}
			}
		}
		Sleep(ptr->m_nDuration);
	}
	ptr->m_pTxThread = 0;

	return 0;
}

UINT RxThread(LPVOID pInfo)
{
	CMainFrame* ptr = (CMainFrame*)pInfo;
	UINT nSegRcvd = 0;
	UINT nSegSent = 0;
	MIB_TCPSTATS  mibTcpStats;
	UINT nCounter = 0;
	DWORD dwLocalInSegs = 0;

	while(m_bContinue)
	{
		if(ptr->m_bRxContinue)
		{
			if(GetTcpStatistics(&mibTcpStats) == NO_ERROR)
			{
				dwLocalInSegs = mibTcpStats.dwInSegs;
				if(dwLocalInSegs>nSegRcvd)
				{
					nSegRcvd = dwLocalInSegs;
					ptr->FlashScrollLock();
				}
				else if(dwLocalInSegs==nSegRcvd) nCounter++;

				if(nCounter==15)
				{
					nCounter=0;
					ptr->ResetIndicators();
				}
			}
		}
		Sleep(ptr->m_nDuration);
	}
	ptr->m_pRxThread = 0;

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
HANDLE CMainFrame::OpenKeyboardDevice()
{
	HANDLE	hDev = NULL;

	if(!DefineDosDevice(DDD_RAW_TARGET_PATH, "Kbd", "\\Device\\KeyboardClass0")) return INVALID_HANDLE_VALUE;

	hDev = CreateFile("\\\\.\\Kbd",GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);

	return hDev;
}


BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	cs.cx = 0;
	cs.cy = 0;
	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;
	cs.dwExStyle |= WS_EX_TOOLWINDOW;
	cs.lpszClass = AfxRegisterWndClass(0);
	return TRUE;
}

void CMainFrame::FlashNumLock()
{
	m_bTxContinue = FALSE;
	
	if(!m_hDev)
	{
		m_bTxContinue = TRUE;
		return;
	}
	
	if(m_bIsForced)
	{
		INPUT input[2];
        ::ZeroMemory(input, sizeof(input));        
        input[0].type = input[1].type = INPUT_KEYBOARD;
        input[0].ki.wVk  = input[1].ki.wVk = VK_NUMLOCK;        
        input[1].ki.dwFlags = KEYEVENTF_KEYUP;
        ::SendInput(2, input, sizeof(INPUT));
		
		m_bTxContinue = TRUE;
		return;
	}

	if(!m_bIsWinNT)
	{
		BYTE byState[256]={0};
		GetKeyboardState(byState);
		BYTE byPrev = byState[VK_NUMLOCK];
		if(byPrev) byState[VK_NUMLOCK] = 0x00;
		else byState[VK_NUMLOCK] = 0x01;
		
		SetKeyboardState(byState);

		Sleep(m_nDuration);

		byState[VK_NUMLOCK] = byPrev;
		SetKeyboardState(byState);

		m_bTxContinue = TRUE;
		return;
	}

	SKeyInfo            InputBuffer;
	SKeyInfo            OutputBuffer;
	UINT				nLedFlagsMask = 0;
	UINT				nNewFlagsMask = 0;
	BOOL				bToggle = FALSE;
	ULONG				nDataLength = sizeof(SKeyInfo);
	ULONG				nReturnedLength = 0;
	int				    i = 0;

	InputBuffer.nUnitId = 0;
	OutputBuffer.nUnitId = 0;

	if(!DeviceIoControl(m_hDev, IOCTL_KEYBOARD_QUERY_INDICATORS, &InputBuffer, nDataLength, &OutputBuffer, nDataLength, &nReturnedLength, NULL))
	{
		m_bTxContinue = TRUE;
		return;
	}

	nLedFlagsMask = (OutputBuffer.nLedFlags);

	switch(nLedFlagsMask)
	{
	case 0:
		nNewFlagsMask = 2;
		break;
	case 1:
		nNewFlagsMask = 3;
		break;
	case 2:
		nNewFlagsMask = 0;
		break;
	case 3:
		nNewFlagsMask = 1;
		break;
	case 4:
		nNewFlagsMask = 6;
		break;
	case 5:
		nNewFlagsMask = 7;
		break;
	case 6:
		nNewFlagsMask = 4;
		break;
	case 7:
		nNewFlagsMask = 5;
		break;
	}
	InputBuffer.nLedFlags = nNewFlagsMask;

	if(!DeviceIoControl(m_hDev, IOCTL_KEYBOARD_SET_INDICATORS,&InputBuffer,nDataLength,NULL,0,&nReturnedLength, NULL))
	{
		m_bTxContinue = TRUE;
		return;
	}

	Sleep(m_nDuration);
	
	InputBuffer.nLedFlags = nLedFlagsMask;
	DeviceIoControl(m_hDev, IOCTL_KEYBOARD_SET_INDICATORS,&InputBuffer,nDataLength,NULL,0,&nReturnedLength, NULL);
	
	m_bTxContinue = TRUE;
}

void CMainFrame::ResetIndicators()
{
	if(!m_bRxContinue || !m_bTxContinue) return;
	
	if(!m_hDev)	return;
	
	if(m_bIsForced)
	{
		INPUT numlock[2];
		::ZeroMemory(numlock, sizeof(numlock));        
		numlock[0].type = numlock[1].type = INPUT_KEYBOARD;
		numlock[0].ki.wVk  = numlock[1].ki.wVk = VK_NUMLOCK;        
		numlock[1].ki.dwFlags = KEYEVENTF_EXTENDEDKEY;
		if(LOWORD(GetKeyState(VK_NUMLOCK))) ::SendInput(2, numlock, sizeof(INPUT));

		INPUT scroll[2];
		::ZeroMemory(scroll, sizeof(scroll));        
		scroll[0].type = scroll[1].type = INPUT_KEYBOARD;
		scroll[0].ki.wVk  = scroll[1].ki.wVk = VK_SCROLL;        
		scroll[1].ki.dwFlags = KEYEVENTF_EXTENDEDKEY;
		if(LOWORD(GetKeyState(VK_SCROLL))) ::SendInput(2, scroll, sizeof(INPUT));
	}
	
	if(!m_bIsWinNT)
	{
		BYTE byState[256]={0};
		GetKeyboardState(byState);
		byState[VK_SCROLL] = 0x00;
		byState[VK_NUMLOCK] = 0x00;
		
		SetKeyboardState(byState);
		return;
	}


	SKeyInfo            InputBuffer;
	SKeyInfo            OutputBuffer;
	UINT				nLedFlagsMask = 0;
	UINT				nNewFlagsMask = 0;
	BOOL				bToggle = FALSE;
	ULONG				nDataLength = sizeof(SKeyInfo);
	ULONG				nReturnedLength = 0;
	int				    i = 0;

	InputBuffer.nUnitId = 0;
	OutputBuffer.nUnitId = 0;

	if(!DeviceIoControl(m_hDev, IOCTL_KEYBOARD_QUERY_INDICATORS, &InputBuffer, nDataLength, &OutputBuffer, nDataLength, &nReturnedLength, NULL)) return;

	nLedFlagsMask = (OutputBuffer.nLedFlags);

	switch(nLedFlagsMask)
	{
	case 0:
		nNewFlagsMask = 0;
		break;
	case 1:
		nNewFlagsMask = 0;
		break;
	case 2:
		nNewFlagsMask = 0;
		break;
	case 3:
		nNewFlagsMask = 0;
		break;
	case 4:
		nNewFlagsMask = 4;
		break;
	case 5:
		nNewFlagsMask = 4;
		break;
	case 6:
		nNewFlagsMask = 4;
		break;
	case 7:
		nNewFlagsMask = 4;
		break;
	}
	InputBuffer.nLedFlags = nNewFlagsMask;

	DeviceIoControl(m_hDev, IOCTL_KEYBOARD_SET_INDICATORS,&InputBuffer,nDataLength,NULL,0,&nReturnedLength, NULL);
}

void CMainFrame::FlashScrollLock()
{
	m_bRxContinue = FALSE;

	if(!m_hDev)
	{
		m_bRxContinue = TRUE;
		return;
	}
	
	if(m_bIsForced)
	{
		INPUT input[2];
		::ZeroMemory(input, sizeof(input));        
		input[0].type = input[1].type = INPUT_KEYBOARD;
		input[0].ki.wVk  = input[1].ki.wVk = VK_SCROLL;        
		input[1].ki.dwFlags = KEYEVENTF_KEYUP;
		::SendInput(2, input, sizeof(INPUT));

		m_bRxContinue = TRUE;
		return;
	}

	if(!m_bIsWinNT)
	{
		BYTE byState[256]={0};
		GetKeyboardState(byState);
		BYTE byPrev = byState[VK_SCROLL];
		if(byPrev) byState[VK_SCROLL] = 0x00;
		else byState[VK_SCROLL] = 0x01;
		

		SetKeyboardState(byState);

		Sleep(m_nDuration);

		byState[VK_SCROLL] = byPrev;
		SetKeyboardState(byState);

		m_bRxContinue = TRUE;
		return;
	}


	SKeyInfo            InputBuffer;
	SKeyInfo            OutputBuffer;
	UINT				nLedFlagsMask = 0;
	UINT				nNewFlagsMask = 0;
	BOOL				bToggle = FALSE;
	ULONG				nDataLength = sizeof(SKeyInfo);
	ULONG				nReturnedLength = 0;
	int				    i = 0;

	InputBuffer.nUnitId = 0;
	OutputBuffer.nUnitId = 0;

	if(!DeviceIoControl(m_hDev, IOCTL_KEYBOARD_QUERY_INDICATORS, &InputBuffer, nDataLength, &OutputBuffer, nDataLength, &nReturnedLength, NULL))
	{
		m_bRxContinue = TRUE;
		return;
	}

	nLedFlagsMask = (OutputBuffer.nLedFlags);

	switch(nLedFlagsMask)
	{
	case 0:
		nNewFlagsMask = 1;
		break;
	case 1:
		nNewFlagsMask = 0;
		break;
	case 2:
		nNewFlagsMask = 3;
		break;
	case 3:
		nNewFlagsMask = 2;
		break;
	case 4:
		nNewFlagsMask = 5;
		break;
	case 5:
		nNewFlagsMask = 4;
		break;
	case 6:
		nNewFlagsMask = 7;
		break;
	case 7:
		nNewFlagsMask = 6;
		break;
	}

	InputBuffer.nLedFlags = nNewFlagsMask;

	if(!DeviceIoControl(m_hDev, IOCTL_KEYBOARD_SET_INDICATORS,&InputBuffer,nDataLength,NULL,0,&nReturnedLength, NULL))
	{
		m_bRxContinue = TRUE;
		return;
	}

	Sleep(m_nDuration);
	
	InputBuffer.nLedFlags = nLedFlagsMask;
	DeviceIoControl(m_hDev, IOCTL_KEYBOARD_SET_INDICATORS,&InputBuffer,nDataLength,NULL,0,&nReturnedLength, NULL);

	m_bRxContinue = TRUE;
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CNewFrameWnd::OnCreate(lpCreateStruct) == -1) return -1;

	HICON hIcon = LoadIcon(lpCreateStruct->hInstance,MAKEINTRESOURCE(IDR_MAINFRAME));
	TrayMessage(m_hWnd, NIM_ADD, 0, hIcon, "Network Lights");

	SetWindowText("Network Lights");

	m_hDev = OpenKeyboardDevice();

	OSVERSIONINFO osvi;
	ZeroMemory(&osvi, sizeof(OSVERSIONINFO));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	if(!GetVersionEx(&osvi)) ZeroMemory(&osvi, sizeof(OSVERSIONINFO));
	
	// Windows95 if:
	// Major == 4 and Minor == 0 and PlatformId != NT
	
	// Windows98 if:
	// Major >= 4 and Minor > 0 and PlatformId != NT
	// (except Major == 4 and Minor == 90 which is ME)
	// (note:  Major == 4 and Minor == 0 is 95)

	// WindowsME if:
	// Major == 4 and Minor == 90 and PlatformId != NT

	// WindowsNT4 if:
	// Major == 4 and Minor == 0 and PlatformId == NT

	// WindowsXP if:
	// Major == 5 and Minor == 1 and PlatformId == NT

	// Windows2000 if:
	// Major == 5 and Minor == 0 and PlatformId == NT

	if(osvi.dwPlatformId == VER_PLATFORM_WIN32_NT && osvi.dwMajorVersion>=4) m_bIsWinNT = TRUE;
	else m_bIsWinNT = FALSE;

	m_bContinue = TRUE;
	GetSettings();

	ResetIndicators();

	if(!m_pRxThread) m_pRxThread = AfxBeginThread((AFX_THREADPROC)RxThread,this, THREAD_PRIORITY_NORMAL,CREATE_SUSPENDED);
	if(!m_pTxThread) m_pTxThread = AfxBeginThread((AFX_THREADPROC)TxThread,this, THREAD_PRIORITY_NORMAL,CREATE_SUSPENDED);

	return 0;
}

void CMainFrame::OnDestroy() 
{
	CNewFrameWnd::OnDestroy();
	
	DefineDosDevice(DDD_REMOVE_DEFINITION, "Kbd", NULL);
	if(m_hDev) CloseHandle(m_hDev);

	m_bContinue = FALSE;
	while(m_pRxThread) Sleep(m_nDuration);
	while(m_pTxThread) Sleep(m_nDuration);
	TrayMessage(m_hWnd, NIM_DELETE, 0, 0, 0);
}

BOOL CMainFrame::TrayMessage(HWND hWnd, DWORD dwMessage, UINT uID, HICON hIcon, PSTR pszTip) 
{ 
    BOOL bResult = FALSE; 

    NOTIFYICONDATA tnd; 

    tnd.cbSize = sizeof(NOTIFYICONDATA); 
    tnd.hWnd = hWnd; 
    tnd.uID = uID; 

    tnd.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP; 
    tnd.uCallbackMessage = WM_ICONTRAY;
    tnd.hIcon = hIcon; 

    if(pszTip) 
    { 
        lstrcpyn(tnd.szTip, pszTip, sizeof(tnd.szTip)); 
    } 
    else 
    { 
        tnd.szTip[0] = '\0'; 
    } 

    bResult = Shell_NotifyIcon(dwMessage, &tnd); 

    return bResult; 
}

LRESULT CMainFrame::OnIconTray(WPARAM wParam, LPARAM lParam)
{
	if(lParam==WM_RBUTTONUP || lParam==WM_LBUTTONUP)
	{
		CPoint pt;
		GetCursorPos(&pt);
		
		CNewMenu popupMenu;
		popupMenu.CreatePopupMenu();
		popupMenu.SetBitmapBackground(RGB(255,0,0));

		CBitmap bmTemp;
		bmTemp.LoadBitmap(IDB_BITMAP0);
		popupMenu.AppendODMenu("About", MF_STRING, ID_ABOUT, &bmTemp);
		bmTemp.DeleteObject();
		bmTemp.LoadBitmap(IDB_BITMAP1);
		popupMenu.AppendODMenu("Settings", MF_STRING, ID_SETTINGS, &bmTemp);
		bmTemp.DeleteObject();
		bmTemp.LoadBitmap(IDB_BITMAP2);
		popupMenu.AppendODMenu("Network Statistics", MF_STRING, ID_STATISTICS, &bmTemp);
		popupMenu.AppendMenu(MF_SEPARATOR);
		bmTemp.DeleteObject();
		bmTemp.LoadBitmap(IDB_BITMAP3);
		popupMenu.AppendODMenu("Exit", MF_STRING, ID_EXIT, &bmTemp);

		popupMenu.TrackPopupMenu(TPM_LEFTALIGN | TPM_LEFTBUTTON,pt.x, pt.y, this);
	}
	return 0;
}

void CMainFrame::GetSettings()
{
	HKEY hKey = NULL;
	DWORD dwType = NULL;
	DWORD dwRes = NULL;
	DWORD dwSize = sizeof(DWORD);
	ULONG que = 0;
	m_bIsForced = FALSE;
	LONG query = RegOpenKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\Network Lights",0,KEY_ALL_ACCESS,&hKey);
	if (query == ERROR_SUCCESS)
	{
		que = RegQueryValueEx(hKey,"Duration",NULL,&dwType,(LPBYTE)&dwRes,&dwSize);
		if(que==ERROR_SUCCESS  && dwRes !=0)
		{
			if(dwRes<100 || dwRes>1000) dwRes = 100;
			m_nDuration = dwRes;
		}

		que = RegQueryValueEx(hKey,"ForceCapture",NULL,&dwType,(LPBYTE)&dwRes,&dwSize);
		if(que==ERROR_SUCCESS  && dwRes !=0) m_bIsForced = TRUE;
	}
	if(hKey) RegCloseKey(hKey);
}

void CMainFrame::OnSettings() 
{
	CSettingsDialog sd;
	if(sd.DoModal()==IDOK) GetSettings();
}

void CMainFrame::OnStatistics() 
{
	CStatisticsDialog sd;
	sd.DoModal();
}

void CMainFrame::OnAbout() 
{
	CAboutDialog ad;
	ad.DoModal();
}

void CMainFrame::OnExit() 
{
	DestroyWindow();
}
